import ij.*;
import ij.process.*;
import ij.gui.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import ij.plugin.*;
import ij.plugin.frame.*;

public class Images_Calculator implements PlugIn, ItemListener {

   MultiLineLabel msg = null;
   TreeMap<String,String> msgs = new TreeMap<String,String>();
   public void run(String arg) {

      GenericDialog gd = new GenericDialog("Images Calculator");
      gd.addMessage("Images");
      gd.addPanel(panel,GridBagConstraints.NORTH, new Insets(0, 0, 0, 0));
      panel.setLayout(new GridBagLayout());
      itemStateChanged(null);

      msgs.put("Sum",       "Sum all above.");
      msgs.put("Mean",      "Algebraic mean all above.");
      msgs.put("SumSq",     "Sum of square of all above.\n ChiSquare");
      msgs.put("MeanSq",    "Algebraic mean of square of all above.");
      msgs.put("SqrtSumSq", "Square Root of Sum of Square of all of above.\nRMS, Magnitude.");
      msgs.put("SumAbs",    "Sum of Absolute of all above.");
      msgs.put("MeanAbs",   "Algebraic mean of absolute of all above.");
      msgs.put("Subtract",  "First minus rest");
      msgs.put("Prod",      "Product of all above.");
      msgs.put("GeoMean",   "Geometric Mean of all of above.");
      msgs.put("Divide",    "Divide first by rest.");
      msgs.put("And",       "Logical AND all above.");
      msgs.put("Or",        "Logical OR all above.");
      msgs.put("Xor",       "Logical Exclusive-or all above.");
      msgs.put("Min",       "Minimum of all above.");
      msgs.put("MinAbs",    "Minumum of absolute of all above.");
      msgs.put("Max",       "Maximum of all above.");
      msgs.put("MaxAbs",    "Maximum of absolute of all above.");
      msgs.put("Phase",     "atan2(first,second)\nrest ignored.");
      msgs.put("DiffPhase", "phase(exp(i*first)/exp(i*second))\nAngle from first to second.\rest ignored.");
      msgs.put("Roll",      "phase(exp(i*first)*exp(i*second))\nFirst advance in phase by second.\nrest ignored.");
      msgs.put("MagPhaReIm","real = first*cos(second);\nimaginary = first*sin(second);\n rest ignored.");
      msgs.put("ReImMagPha","magnitude = SqrtSumSq(first,second);\nphase = Phase(first,second);\nrest ignored.");
      msgs.put("RollRI",    "Roll complex(first:real,second:imaginary) by third:phase\nrest ignored.");
      msgs.put("SumMagPha", "Complex algebraic sum of pairs of above.\nFirst convert to re im.");
      msgs.put("MeanMagPha","Complex algebraic mean of pairs of above.\nFirst convert to re im.");
      msgs.put("SubMagPha", "Complex subtraction of pairs of above.\nFirst convert to re im.");
      msgs.put("FFT",       "Complex forward FFT.");
      msgs.put("Label",     "Set Slice Label.\nSame label on all slices of each ImagePlus.");
      msgs.put("LabelFrames","Set Slice Label.\nSame label on all z slices of each frame.");
      msgs.put("Stack",     "Create stack from all above.\nAppend all slices in Z dimension.");
      msgs.put("HyperStack","Create hyperstack from all above.\nAppend slices and set Z dimension same as first.");
      msgs.put("HyperSplit","Extract each frame.");

      gd.addChoice("Operation:", new String[]{"Sum", "Mean", "SumSq", "MeanSq", "SqrtSumSq", "SqrtMeanSq", "SumAbs", "MeanAbs", "Subtract",
                                              "Prod", "GeoMean", "Divide",
                                              "And", "Or", "Xor", "Min", "MinAbs", "Max", "MaxAbs",
                                              "Phase", "DiffPhase", "Roll",
                                              "MagPhaReIm", "ReImMagPha", "RollRI", "SumMagPha", "MeanMagPha", "SubMagPha",
                                              "FFT",
                                              "Label", "LabelFrames", "Stack", "HyperStack", "HyperSplit"}, "");

      gd.addMessage("Choose operation\n ");
      msg = (MultiLineLabel)gd.getMessage();
      msg.setText(msgs.get("Sum"));

      ((Choice)gd.getChoices().get(0)).addItemListener(new ItemListener() {
         @Override public void itemStateChanged(ItemEvent e) {
            msg.setText(msgs.get(((Choice)e.getSource()).getSelectedItem()));
            gd.pack();
            };
         });

      gd.showDialog();
      if (gd.wasCanceled())
         return;

      String cmd = gd.getNextChoice();
      ImagePlus[] ress = null;
      switch (cmd) {
         case "MagPhaReIm": ress = magphareim(getChoices()); break;
         case "ReImMagPha": ress = reimmagpha(getChoices()); break;
         case "RollRI":     ress = rollRI(getChoices());     break;
         case "SumMagPha":  ress = sumMagPha(getChoices());  break;
         case "MeanMagPha": ress = meanMagPha(getChoices()); break;
         case "SubMagPha":  ress = subMagPha(getChoices());  break;
         case "HyperSplit": ress = hyperSplit(getChoices()); break;
         case "LabelFrames": ress = labelFrames(getChoices()); break;
         default:           ress = new ImagePlus[]{process(cmd, getChoices())};
         }
      for(ImagePlus res : ress)
         if (res != null) {
            res.changes = false;
            res.show();
            }
      //ImagePlus res =  process(gd.getNextChoice(), getChoices());
      //if (res != null) res.show();
      }

   public static String[] re2names(String re) {
      return re2names(re, WindowManager.getImageTitles());
      }
   public static String[] re2names(String re, String[] its) {
      Vector<String> vs = new Vector<String>();
      for(String n : its)
         if (n.matches(re))
            vs.add(n);
      return vs.toArray(new String[0]);
      }
   public static ImagePlus[] titles2imps(String ... strs) {
      ImagePlus[] imps = new ImagePlus[strs.length];
      for(int i=0; i<strs.length; i++)
         imps[i] = WindowManager.getImage(strs[i]);
      return imps;
      }
   public static ImagePlus process(String command, String ... strs) {
      return process(command, titles2imps(strs));
      }
   public static ImagePlus process(String cmd, ImagePlus ... imps) {
      switch (cmd) {
         case "Sum":       return sum(imps);
         case "Mean":      return mean(imps);
         case "SumSq":     return sumSq(imps);
         case "MeanSq":    return meanSq(imps);
         case "SqrtSumSq": return sqrtSumSq(imps);
         case "SqrtMeanSq":return sqrtMeanSq(imps);
         case "SumAbs":    return sumAbs(imps);
         case "MeanAbs":   return meanAbs(imps);
         case "Subtract":  return sub(imps);
         case "Prod":      return prod(imps);
         case "GeoMean":   return geoMean(imps);
         case "Divide":    return div(imps);
         case "And":       return and(imps);
         case "Or":        return or(imps);
         case "Xor":       return xor(imps);
         case "Min":       return min(imps);
         case "MinAbs":    return minAbs(imps);
         case "Max":       return max(imps);
         case "MaxAbs":    return maxAbs(imps);
         case "Phase":     return phase(imps);
         case "DiffPhase": return diffPhase(imps);
         case "Roll":      return roll(imps);
         case "FFT":       return FFT(imps);
         case "Label":     return label(imps);
         case "Stack":     return stack(imps);
         case "HyperStack":return hyperStack(imps);
         }
      return null;
      }
   public static ImagePlus sum(String ... strs) {
      return sum(titles2imps(strs));
      }
   public static ImagePlus sum(ImagePlus ... imps) {
      if (imps == null || imps.length == 0) return null;
      ImagePlus res = imps[0].duplicate();
      res.setTitle("Sum "+imps[0].getTitle());
      for(int i=1; i<imps.length; i++)
         res = ImageCalculator.run(res, imps[i], "add 32-bit stack");
      return res;
      }
   public static ImagePlus mean(String ... strs) {
      return mean(titles2imps(strs));
      }
   public static ImagePlus mean(ImagePlus ... imps) {
      ImagePlus res = sum(imps);
      res.setTitle("Mean "+imps[0].getTitle());
      IJ.run(res, "Divide...", "value="+imps.length+" stack");
      return res;
      }
   public static ImagePlus sumSq(String ... strs) {
      return sumSq(titles2imps(strs));
      }
   public static ImagePlus sumSq(ImagePlus ... imps) {
      if (imps == null || imps.length == 0) return null;
      ImagePlus res = imps[0].duplicate();
      res.setTitle("SumSq "+imps[0].getTitle());
      IJ.run(res,"Square","stack");
      for(int i=1; i<imps.length; i++) {
         ImagePlus tmp = imps[i].duplicate();
         IJ.run(tmp,"Square","stack");
         res = ImageCalculator.run(res, tmp, "add 32-bit stack");
         }
      return res;
      }
   public static ImagePlus meanSq(String ... strs) {
      return meanSq(titles2imps(strs));
      }
   public static ImagePlus meanSq(ImagePlus ... imps) {
      ImagePlus res = sumSq(imps);
      IJ.run(res, "Divide...", "value="+imps.length+" stack");
      res.setTitle("MeanSq "+imps[0].getTitle());
      return res;
      }
   public static ImagePlus sqrtSumSq(String ... strs) {
      return sqrtSumSq(titles2imps(strs));
      }
   public static ImagePlus sqrtSumSq(ImagePlus ... imps) {
      ImagePlus res = sumSq(imps);
      res.setTitle("SqrtSumSq "+imps[0].getTitle());
      IJ.run(res, "Square Root", "stack");
      return res;
      }
   public static ImagePlus sqrtMeanSq(String ... strs) {
      return sqrtMeanSq(titles2imps(strs));
      }
   public static ImagePlus sqrtMeanSq(ImagePlus ... imps) {
      ImagePlus res = meanSq(imps);
      res.setTitle("SqrtMeanSq "+imps[0].getTitle());
      IJ.run(res, "Square Root", "stack");
      return res;
      }
   public static ImagePlus sumAbs(String ... strs) {
      return sumAbs(titles2imps(strs));
      }
   public static ImagePlus sumAbs(ImagePlus ... imps) {
      if (imps == null || imps.length == 0) return null;
      ImagePlus res = imps[0].duplicate();
      res.setTitle("SumAbs "+imps[0].getTitle());
      IJ.run(res,"Abs","stack");
      for(int i=1; i<imps.length; i++) {
         ImagePlus tmp = imps[i].duplicate();
         IJ.run(tmp,"Abs","stack");
         res = ImageCalculator.run(res, tmp, "add 32-bit stack");
         }
      return res;
      }
   public static ImagePlus meanAbs(String ... strs) {
      return meanAbs(titles2imps(strs));
      }
   public static ImagePlus meanAbs(ImagePlus ... imps) {
      ImagePlus res = sumAbs(imps);
      res.setTitle("MeanAbs "+imps[0].getTitle());
      IJ.run(res, "Divide...", "value="+imps.length+" stack");
      return res;
      }
   public static ImagePlus sub(String ... strs) {
      return sub(titles2imps(strs));
      }
   public static ImagePlus sub(ImagePlus ... imps) {
      if (imps == null || imps.length == 0) return null;
      ImagePlus res = imps[0].duplicate();
      res.setTitle("Sub "+imps[0].getTitle());
      for(int i=1; i<imps.length; i++)
         res = ImageCalculator.run(res, imps[i], "Subtract 32-bit stack");
      return res;
      }
   public static ImagePlus prod(String ... strs) {
      return prod(titles2imps(strs));
      }
   public static ImagePlus prod(ImagePlus ... imps) {
      if (imps == null || imps.length == 0) return null;
      ImagePlus res = imps[0].duplicate();
      res.setTitle("Prod "+imps[0].getTitle());
      for(int i=1; i<imps.length; i++)
         res = ImageCalculator.run(res, imps[i], "multiply 32-bit stack");
      return res;
      }
   public static ImagePlus geoMean(String ... strs) {
      return geoMean(titles2imps(strs));
      }
   public static ImagePlus geoMean(ImagePlus ... imps) {
      if (imps == null || imps.length == 0) return null;
      ImagePlus res = imps[0].duplicate();
      res.setTitle("GeoMean "+imps[0].getTitle());
      IJ.run(res, "Log", "stack");
      IJ.run(res, "Divide...", "value="+imps.length+" stack");
      IJ.run(res, "Exp", "stack");
      for(int i=1; i<imps.length; i++){
         ImagePlus ughimp = imps[i].duplicate();
         IJ.run(ughimp, "Log", "stack");
         IJ.run(ughimp, "Divide...", "value="+imps.length+" stack");
         IJ.run(ughimp, "Exp", "stack");
         res = ImageCalculator.run(res, ughimp, "Multiply 32-bit stack");
         }
      return res;
      }
   public static ImagePlus div(String ... strs) {
      return sub(titles2imps(strs));
      }
   public static ImagePlus div(ImagePlus ... imps) {
      if (imps == null || imps.length == 0) return null;
      ImagePlus res = imps[0].duplicate();
      res.setTitle("Div "+imps[0].getTitle());
      for(int i=1; i<imps.length; i++)
         res = ImageCalculator.run(res, imps[i], "Divide 32-bit stack");
      return res;
      }
   public static ImagePlus and(String ... strs) {
      return and(titles2imps(strs));
      }
   public static ImagePlus and(ImagePlus ... imps) {
      if (imps == null || imps.length == 0) return null;
      ImagePlus res = imps[0].duplicate();
      res.setTitle("And "+imps[0].getTitle());
      for(int i=1; i<imps.length; i++)
         res = ImageCalculator.run(res, imps[i], "AND 32-bit stack");
      return res;
      }
   public static ImagePlus or(String ... strs) {
      return or(titles2imps(strs));
      }
   public static ImagePlus or(ImagePlus ... imps) {
      if (imps == null || imps.length == 0) return null;
      ImagePlus res = imps[0].duplicate();
      res.setTitle("Or "+imps[0].getTitle());
      for(int i=1; i<imps.length; i++)
         res = ImageCalculator.run(res, imps[i], "OR 32-bit stack");
      return res;
      }
   public static ImagePlus xor(String ... strs) {
      return xor(titles2imps(strs));
      }
   public static ImagePlus xor(ImagePlus ... imps) {
      if (imps == null || imps.length == 0) return null;
      ImagePlus res = imps[0].duplicate();
      res.setTitle("Xor "+imps[0].getTitle());
      for(int i=1; i<imps.length; i++)
         res = ImageCalculator.run(res, imps[i], "XOR 32-bit stack");
      return res;
      }
   public static ImagePlus min(String ... strs) {
      return min(titles2imps(strs));
      }
   public static ImagePlus min(ImagePlus ... imps) {
      if (imps == null || imps.length == 0) return null;
      ImagePlus res = imps[0].duplicate();
      res.setTitle("Min "+imps[0].getTitle());
      for(int i=1; i<imps.length; i++)
         res = ImageCalculator.run(res, imps[i], "Min 32-bit stack");
      return res;
      }
   public static ImagePlus minAbs(String ... strs) {
      return minAbs(titles2imps(strs));
      }
   public static ImagePlus minAbs(ImagePlus ... imps) {
      if (imps == null || imps.length == 0) return null;
      ImagePlus res = imps[0].duplicate();
      res.setTitle("MinAbs "+imps[0].getTitle());
      IJ.run(res,"Abs","stack");
      for(int i=1; i<imps.length; i++) {
         ImagePlus tmp = imps[i].duplicate();
         IJ.run(tmp,"Abs","stack");
         res = ImageCalculator.run(res, tmp, "Min 32-bit stack");
         }
      return res;
      }
   public static ImagePlus max(String ... strs) {
      return max(titles2imps(strs));
      }
   public static ImagePlus max(ImagePlus ... imps) {
      if (imps == null || imps.length == 0) return null;
      ImagePlus res = imps[0].duplicate();
      res.setTitle("Max "+imps[0].getTitle());
      for(int i=1; i<imps.length; i++)
         res = ImageCalculator.run(res, imps[i], "Max 32-bit stack");
      return res;
      }
   public static ImagePlus maxAbs(String ... strs) {
      return maxAbs(titles2imps(strs));
      }
   public static ImagePlus maxAbs(ImagePlus ... imps) {
      if (imps == null || imps.length == 0) return null;
      ImagePlus res = imps[0].duplicate();
      res.setTitle("MaxAbs "+imps[0].getTitle());
      IJ.run(res,"Abs","stack");
      for(int i=1; i<imps.length; i++) {
         ImagePlus tmp = imps[i].duplicate();
         IJ.run(tmp,"Abs","stack");
         res = ImageCalculator.run(res, tmp, "Max 32-bit stack");
         }
      return res;
      }

   public static ImagePlus phase(String ... strs) {
      return phase(titles2imps(strs));
      }
   public static ImagePlus phase(ImagePlus ... imps) {
      if (imps == null || imps.length < 2) return null;
      ImagePlus res = imps[0].duplicate();
      res.setTitle("Phase "+imps[0].getTitle());
      int ns = imps[0].getImageStackSize();
      for(int s=0; s<ns; s++) {
         float[][] re = imps[0].getStack().getProcessor(s+1).getFloatArray();
         float[][] im = imps[1].getStack().getProcessor(s+1).getFloatArray();
         float[][] ph = res.getStack().getProcessor(s+1).getFloatArray();
         for(int ix=0; ix<re.length; ix++) 
            for(int iy=0; iy<re[0].length; iy++) 
               ph[ix][iy] = (float)Math.atan2(im[ix][iy],re[ix][iy]);
         res.getStack().getProcessor(s+1).setFloatArray(ph);
         }
      return res;
      }
   public static ImagePlus diffPhase(String ... strs) {
      return diffPhase(titles2imps(strs));
      }
   public static ImagePlus diffPhase(ImagePlus ... imps) {
      if (imps == null || imps.length < 2) return null;
      ImagePlus res = imps[0].duplicate();
      res.setTitle("Phase "+imps[0].getTitle());
      int ns = imps[0].getImageStackSize();
      for(int s=0; s<ns; s++) {
         float[][] p1 = imps[0].getStack().getProcessor(s+1).getFloatArray();
         float[][] p2 = imps[1].getStack().getProcessor(imps[1].getStackSize()==1?1:s+1).getFloatArray();
         float[][] pd = res.getStack().getProcessor(s+1).getFloatArray();
         for(int ix=0; ix<p1.length; ix++) 
            for(int iy=0; iy<p1[0].length; iy++) 
               pd[ix][iy] = diffphase(p1[ix][iy],p2[ix][iy]);
         res.getStack().getProcessor(s+1).setFloatArray(pd);
         }
      return res;
      }
public static float diffphase(float p1, float p2) {
   float r1 = (float) Math.cos(p1);
   float i1 = (float) Math.sin(p1);
   float r2 = (float) Math.cos(p2);
   float i2 = (float)-Math.sin(p2);
   return diffphase(r1,i1,r2,i2);
   }
public static float[] diffphase(float r1, float i1, float p) {
   float r2 = (float) Math.cos(p);
   float i2 = (float)-Math.sin(p);
   float r3 = r1*r2 + -1*i1*i2;
   float i3 = r1*i2 +    r2*i1;
   return new float[]{r3,i3};
   }
public static float diffphase(float r1, float i1, float r2, float i2) {
   float r3 = r1*r2 + -1*i1*i2;
   float i3 = r1*i2 +    r2*i1;
   return (float)(Math.atan2(i3,r3));
   }
   public static ImagePlus roll(String ... strs) {
      return roll(titles2imps(strs));
      }
   public static ImagePlus roll(ImagePlus ... imps) {
      if (imps == null || imps.length < 2) return null;
      ImagePlus res = imps[0].duplicate();
      res.setTitle("Phase "+imps[0].getTitle());
      int ns = imps[0].getImageStackSize();
      for(int s=0; s<ns; s++) {
         float[][] p1 = imps[0].getStack().getProcessor(s+1).getFloatArray();
         float[][] p2 = imps[1].getStack().getProcessor(imps[1].getStackSize()==1?1:s+1).getFloatArray();
         float[][] pd = res.getStack().getProcessor(s+1).getFloatArray();
         for(int ix=0; ix<p1.length; ix++) 
            for(int iy=0; iy<p1[0].length; iy++) 
               pd[ix][iy] = diffphase(p1[ix][iy],-p2[ix][iy]);
         res.getStack().getProcessor(s+1).setFloatArray(pd);
         }
      return res;
      }
   public static ImagePlus[] rollRI(String ... strs) {
      return rollRI(titles2imps(strs));
      }
   public static ImagePlus[] rollRI(ImagePlus ... imps) {
      if (imps == null || imps.length < 3) return null;
      ImagePlus resR = imps[0].duplicate();
      ImagePlus resI = imps[1].duplicate();
      resR.setTitle("Rolled real "+imps[0].getTitle());
      resI.setTitle("Rolled imaginary "+imps[1].getTitle());
      int ns = imps[0].getNSlices();
      int nf = imps[0].getNFrames();
      for(int f=0; f<nf; f++) {
         for(int s=0; s<ns; s++) {
            int si = imps[0].getStackIndex(1,s+1,f+1);
            int sp = imps[2].getStackIndex(1,s+1,Math.min(f+1,imps[2].getNFrames()));
            float[][] re = imps[0].getStack().getProcessor(si).getFloatArray();
            float[][] im = imps[1].getStack().getProcessor(si).getFloatArray();
            float[][] pa = imps[2].getStack().getProcessor(sp).getFloatArray();
            //float[][] pd = res.getStack().getProcessor(si).getFloatArray();
            float[][] rr = resR.getStack().getProcessor(si).getFloatArray();
            float[][] ri = resI.getStack().getProcessor(si).getFloatArray();
            for(int ix=0; ix<re.length; ix++) 
               for(int iy=0; iy<re[0].length; iy++) {
                  float[] r = diffphase(re[ix][iy],im[ix][iy],pa[ix][iy]);
                  rr[ix][iy] = r[0];
                  ri[ix][iy] = r[1];
                  }
            resR.getStack().getProcessor(si).setFloatArray(rr);
            resI.getStack().getProcessor(si).setFloatArray(ri);
            }
         }
      return new ImagePlus[]{resR,resI};
      }

   public static ImagePlus[] magphareim(ImagePlus ... imps) {
      if (imps == null || imps.length < 2) return null;
      ImagePlus re = imps[0].duplicate();
      ImagePlus im = imps[1].duplicate();
      re.setTitle("Real "+imps[0].getTitle());
      im.setTitle("Imaginary "+imps[0].getTitle());
      int ns = imps[0].getImageStackSize();
      for(int s=0; s<ns; s++) {
         float[][] pm = imps[0].getStack().getProcessor(s+1).getFloatArray();
         float[][] pp = imps[1].getStack().getProcessor(s+1).getFloatArray();
         float[][] pr = re.getStack().getProcessor(s+1).getFloatArray();
         float[][] pi = im.getStack().getProcessor(s+1).getFloatArray();
         for(int ix=0; ix<pm.length; ix++) 
            for(int iy=0; iy<pm[0].length; iy++) {
               pr[ix][iy] = pm[ix][iy]*(float)Math.cos(pp[ix][iy]);
               pi[ix][iy] = pm[ix][iy]*(float)Math.sin(pp[ix][iy]);
               }
         re.getStack().getProcessor(s+1).setFloatArray(pr);
         im.getStack().getProcessor(s+1).setFloatArray(pi);
         }
      return new ImagePlus[]{re,im};
      }
   public static ImagePlus[] reimmagpha(ImagePlus ... imps) {
      return new ImagePlus[]{sqrtSumSq(imps),phase(imps)};
      }
   public static ImagePlus[] sumMagPha(ImagePlus ... imps) {
      if (imps.length < 2 || imps.length%2 != 0) return null;
      int n = imps.length/2;
      ImagePlus[] reimps = new ImagePlus[n];
      ImagePlus[] imimps = new ImagePlus[n];
      for(int i=0; i < imps.length; i+=2) {
         ImagePlus[] reim = magphareim(imps[i],imps[i+1]);
         reimps[i/2] = reim[0];
         imimps[i/2] = reim[1];
         }
      return reimmagpha(sum(reimps),sum(imimps));
      }
   public static ImagePlus[] meanMagPha(ImagePlus ... imps) {
      if (imps.length < 2 || imps.length%2 != 0) return null;
      int n = imps.length/2;
      ImagePlus[] reimps = new ImagePlus[n];
      ImagePlus[] imimps = new ImagePlus[n];
      for(int i=0; i < imps.length; i+=2) {
         ImagePlus[] reim = magphareim(imps[i],imps[i+1]);
         reimps[i/2] = reim[0];
         imimps[i/2] = reim[1];
         }
      return reimmagpha(mean(reimps),mean(imimps));
      }
   public static ImagePlus[] subMagPha(ImagePlus ... imps) {
      if (imps.length < 2 || imps.length%2 != 0) return null;
      int n = imps.length/2;
      ImagePlus[] reimps = new ImagePlus[n];
      ImagePlus[] imimps = new ImagePlus[n];
      for(int i=0; i < imps.length; i+=2) {
         ImagePlus[] reim = magphareim(imps[i],imps[i+1]);
         reimps[i/2] = reim[0];
         imimps[i/2] = reim[1];
         }
      return reimmagpha(sub(reimps),sub(imimps));
      }

   public static ImagePlus FFT(ImagePlus ... imps) {
IJ.log("imps "+imps.length);
      if (imps.length==1 && imps[0].getStackSize()==2) {
         IJ.log("FFT");
         return FFT.forward(imps[0]);
         }
      return null;
      }


   public static ImagePlus stack(String ... strs) {
      return stack(titles2imps(strs));
      }
   public static ImagePlus stack(ImagePlus ... imps) {
      if (imps == null || imps.length == 0) return null;
      ImagePlus res = imps[0].duplicate();
      ImageStack is = res.getStack();
      for(int i=1; i<imps.length; i++) {
         ImageStack isi = imps[i].getStack();
         int ns = imps[i].getImageStackSize();
         for(int s=0; s<ns; s++) 
            is.addSlice(isi.getSliceLabel(s+1),isi.getProcessor(s+1));
         }
      res.setStack(is);
      return res;
      }
   public static ImagePlus[] labelFrames(String ... strs) {
      return labelFrames(titles2imps(strs));
      }
   public static ImagePlus[] labelFrames(ImagePlus ... imps) {
      for(int i=0; i<imps.length; i++) {
         GenericDialog gd = new NonBlockingGenericDialog("Labels");
         gd.addMessage(imps[i].getTitle());
         int nf = imps[i].getNFrames();
         for(int f=0; f<nf; f++) {
            String str = imps[i].getStack().getSliceLabel(1)!= null ? imps[i].getStack().getSliceLabel(1).split("\n",2)[0] : "f "+f;
            gd.addStringField("Frame "+f+":", str, 40);
            }
         gd.showDialog();
         if (gd.wasCanceled()) continue;
         String[] labels = new String[nf];
         for(int f=0; f<nf; f++)
            labels[f] = gd.getNextString();
         labelFrames(labels,imps[i]);
         }
      return imps;
      }
   public static ImagePlus labelFrames(String[] labels, ImagePlus imp) {
      int nz = imp.getNSlices();
      int nf = imp.getNFrames();
      for(int f=0; f<nf; f++)
         for(int z=0; z<nz; z++) {
            int i = imp.getStackIndex(1,z+1,f+1);
            String sl = imp.getStack().getSliceLabel(i);
            imp.getStack().setSliceLabel(sl==null ? labels[f] : sl.replaceFirst("^[^\\n]*\\n",labels[f]+"\n"), i);
            }
      imp.repaintWindow();
      return imp;
      }
   public static ImagePlus label(ImagePlus ... imps) {
      GenericDialog gd = new NonBlockingGenericDialog("Labels");
      for(int i=0; i<imps.length; i++) {
         String str = imps[i].getStack().getSliceLabel(1)!= null ? imps[i].getStack().getSliceLabel(1).split("\\n",2)[0] : imps[i].getTitle();
         gd.addStringField(imps[i].getTitle(),str,40);
         }
      gd.showDialog();
      if (gd.wasCanceled()) return null;
      String[] labels = new String[imps.length];
      for(int i=0; i<imps.length; i++)
         labels[i] = gd.getNextString();
      return label(labels,imps);
      }
   public static ImagePlus label(String[] labels, ImagePlus[] imps) {
      if (labels == null || imps == null || labels.length != imps.length) return null;
      for(int i=0; i<labels.length; i++) 
         for(int s=0; s<imps[i].getStackSize(); s++) {
            String label = imps[i].getStack().getSliceLabel(s+1);
            imps[i].getStack().setSliceLabel(label==null ? labels[i] : label.replaceFirst("^[^\\n]*\\n",labels[i]+"\\\n"), s+1);
            imps[i].repaintWindow();
            }
      return null;
      }
   public static ImagePlus hyperStack(String ... strs) {
      return hyperStack(titles2imps(strs));
      }
   public static ImagePlus hyperStack(ImagePlus ... imps) {
      if (imps == null || imps.length == 0) return null;
      ImagePlus res = stack(imps);
      res.setDimensions(imps[0].getNChannels(),imps[0].getNSlices(),res.getImageStackSize()/imps[0].getNSlices());
      res.setOpenAsHyperStack(true);
      return res;
      }
   public static ImagePlus[] hyperSplit(String ... strs) {
      return hyperSplit(titles2imps(strs));
      }
   public static ImagePlus[] hyperSplit(ImagePlus ... imps) {
      if (imps == null) return null;
      Vector<ImagePlus> vri = new Vector<ImagePlus>();
      for(int i=0; i<imps.length; i++) {
         int nc = imps[i].getNChannels();
         int nz = imps[i].getNSlices();
         int nf = imps[i].getNFrames();
         for(int f=0; f<nf; f++)
            vri.add(subHyperStack(imps[i],"1-"+nc,"1-"+nz,""+(f+1)));
         }
      return vri.toArray(new ImagePlus[0]);
      }
   /** and copy the metadata also.
    */
   public static ImagePlus subHyperStack(ImagePlus imp, String cl, String zl, String fl) {
      ImagePlus simp = SubHyperstackMaker.makeSubhyperstack(imp,cl,zl,fl);
      simp.setTitle(imp.getTitle());
      simp.setCalibration(imp.getCalibration());
      simp.setFileInfo(imp.getOriginalFileInfo());
      simp.setProperty("Info",imp.getProperty("Info"));
      simp.setDisplayRange(imp.getDisplayRangeMin(),imp.getDisplayRangeMax());
      return simp;
      }

   Vector<Choice> ico = new Vector<Choice>();
   Panel panel = new Panel();
   public void itemStateChanged(ItemEvent e) {
      Choice c = null;

      GridBagConstraints gbc = new GridBagConstraints();
      gbc.fill = GridBagConstraints.VERTICAL;
      gbc.gridx = 0;
      gbc.gridy = 0;
      gbc.weightx = 0.5;

      if (e == null) {
         c = impChoice(IJ.getImage());
         panel.add(c,gbc);
         ico.add(c);
         c = null;
         }
      else c = (Choice)e.getSource();

      if (c == null || ( c.getSelectedIndex() > 0 && ((Choice)e.getSource()) == panel.getComponent(panel.getComponentCount()-1) )) {
         Choice cc = impChoice(null);
         gbc.gridx = 0;
         gbc.gridy = panel.getComponentCount()+1;
         panel.add(cc,gbc);
         ico.add(c);
         }

      if (e != null) ((GenericDialog)panel.getParent()).pack();
      }

   private ImagePlus[] getChoices() {
      Vector<ImagePlus> imps = new Vector<ImagePlus>();
      for(Choice c : ico)
         if (c != null && c.getSelectedIndex() > 0)
            imps.add(WindowManager.getImage(imgIds[c.getSelectedIndex()-1]));
      return imps.toArray(new ImagePlus[0]);
      }

   int[] imgIds = WindowManager.getIDList();
   private Choice impChoice(ImagePlus imp) {
      Choice c = new Choice();
      c.add("<none>");
      c.select(0);
      for(int i=0; i<imgIds.length; i++) {
         ImagePlus tmpimp = WindowManager.getImage(imgIds[i]);
         c.add(tmpimp.getTitle());
         if (imp!=null && tmpimp.getID() == imp.getID())
            c.select(i+1);
         }
      c.addItemListener(this);
      return c;
      }
}
