import ij.*;
import ij.process.*;
import ij.gui.*;
import java.awt.*;
import java.util.*;
import ij.plugin.*;
import ij.plugin.frame.*;
import ij.measure.ResultsTable;

public class The_Combiner implements PlugIn {

   public void run(String arg) {
      boolean convert2RGB = false;

      if (arg.equals("")) {
         GenericDialog gd = new GenericDialog("The Combiner...");
         gd.setInsets(0, 20, 0);
         gd.addMessage("Choose...");
         CheckboxGroup cbg = new CheckboxGroup();
         int cb = 0;

         gd.setInsets(0, 35, 0);
         gd.addCheckbox("Frame Montage", false);
         ((Checkbox)gd.getCheckboxes().get(cb++)).setCheckboxGroup(cbg);

         gd.setInsets(0, 35, 0);
         gd.addCheckbox("Images Montage", false);
         ((Checkbox)gd.getCheckboxes().get(cb++)).setCheckboxGroup(cbg);

         gd.setInsets(0, 35, 0);
         gd.addCheckbox("Images Combine", false);
         ((Checkbox)gd.getCheckboxes().get(cb++)).setCheckboxGroup(cbg);

         gd.addStringField("RxC:", WindowManager.getWindow("matrix") == null ? "4x2" : "matrix");

         gd.showDialog();
         if (gd.wasCanceled())
            return;
         if (gd.getNextBoolean()) arg = "Frame Montage";
         if (gd.getNextBoolean()) arg = "Images Montage "+gd.getNextString();
         if (gd.getNextBoolean()) arg = "Images Combine "+gd.getNextString();
         }
IJ.log("arg = '"+arg+"'");
      if (arg.equals("Frame Montage")) {
         GenericDialog gd = new GenericDialog("The Frame Combiner...");
         Vector<String> itv = new Vector<String>();
         itv.add("<choose>");
         int[] ids = WindowManager.getIDList();
         if (ids != null)
            for(int id : ids)
               itv.add(WindowManager.getImage(id).getTitle());
         gd.addChoice("Multiframe Image:", itv.toArray(new String[0]), "");
         gd.addStringField("Row Variable","");
         gd.addCheckbox("Convert to RGB", convert2RGB);
         gd.showDialog();
         if (gd.wasCanceled())
            return;
         int ind = gd.getNextChoiceIndex();
         ImagePlus imp = (ind > 0) ? WindowManager.getImage(ids[ind-1]) : IJ.getImage();
         String var = gd.getNextString();
         convert2RGB = gd.getNextBoolean();
         if (convert2RGB) montageWL(imp,var).show();
                     else montage(imp,var).show();
         }
      if (arg.startsWith("Images")) {
         GenericDialog gd = new GenericDialog(arg);
         Panel panel = new Panel();
         panel.setLayout(new GridBagLayout());
         gd.addPanel(panel,GridBagConstraints.NORTH, new Insets(0, 0, 0, 0));

         int[] ids = WindowManager.getIDList();
         Vector<String> itv = new Vector<String>();
         itv.add("<blank>");
         if (ids != null)
            for(int id : ids)
               itv.add(WindowManager.getImage(id).getTitle());

         ResultsTable rt = null;
         int n, m;
         String p = arg.split(" ",3)[2];
         if (p.matches("\\d+x\\d+")) {
            String[] nxm = p.split("x");
            n = Integer.parseInt(nxm[0]);
            m = Integer.parseInt(nxm[1]);
            }
         else {
            rt = ResultsTable.getResultsTable(p);
            if (rt == null) { IJ.error("table '"+p+"' not found"); return; }
            n = rt.size();
            m = rt.getLastColumn()+1;
            }
         Choice[][] matrix = new Choice[n][m];
         for(int i=0; i<n; i++)
            for(int j=0; j<m; j++) {
               matrix[i][j] = new Choice();
               for(String title : itv)
                  matrix[i][j].add(title);
               if (rt != null) 
                  matrix[i][j].select(rt.getStringValue(j,i));
               GridBagConstraints c = new GridBagConstraints();
               c.fill = GridBagConstraints.VERTICAL;
               c.gridx = j;
               c.gridy = i;
               c.weightx = 0.5;
               panel.add(matrix[i][j],c);
               }

         gd.setInsets(5, 35, 0);
         gd.addCheckbox("Convert to RGB", convert2RGB);
         gd.addCheckbox("Currect slice only", true);
         gd.addChoice("Where",new String[]{"Left","Center","Right"}, "Left");
         gd.addChoice("Where",new String[]{"Top","Center","Bottom"}, "Top");

         RoiManager roiManager = RoiManager.getInstance();
         Roi[] rois = null;;
         if (roiManager != null) {
            rois = roiManager.getRoisAsArray();
            Vector<String> rtv = new Vector<String>();
            rtv.add("<none>");
            for(Roi roi : rois)
               rtv.add(roi.getName());
            gd.addChoice("Roi", rtv.toArray(new String[0]),"");
            }

         gd.showDialog();
         if (gd.wasCanceled()) return;

         convert2RGB = gd.getNextBoolean();
         boolean cso = gd.getNextBoolean();
         int where = gd.getNextChoiceIndex() + 3*gd.getNextChoiceIndex();

         Roi roi = null;
         if (rois != null) {
            int ri = gd.getNextChoiceIndex();
            if (ri > 0)
               roi = rois[ri-1];
            }

         rt = new ResultsTable(1);
         ImagePlus[][] images = new ImagePlus[n][m];
         for(int i=0; i<n; i++, rt.incrementCounter())
            for(int j=0; j<m; j++)
               if (matrix[i][j].getSelectedIndex() > 0) {
                  int ind = matrix[i][j].getSelectedIndex()-1;
                  if (ind < 0) continue;
                  ImagePlus imp = WindowManager.getImage(ids[ind]);
                  rt.addValue(""+j,imp.getTitle());
                  images[i][j] = cso ? (new Duplicator()).run(imp,imp.getChannel(),imp.getChannel(),imp.getSlice(),imp.getSlice(),imp.getFrame(),imp.getFrame()) : imp.duplicate();
                  if (roi != null) {
                     images[i][j].setRoi(roi);
                     IJ.run(images[i][j], "Clear Outside", "");
                     images[i][j] = images[i][j].crop();
                     }
                  if (convert2RGB) IJ.run(images[i][j], "RGB Color", "");
                  }
               else rt.addValue(""+j,"");
         rt.deleteRow(rt.size()-1);
         rt.show("matrix");
//IJ.log("arg="+(arg.split(" "))[1]);
              if ((arg.split(" "))[1].equals("Montage")) montage(images,where).show();
         else if ((arg.split(" "))[1].equals("Combine")) combine(images,where).show();
         else IJ.error("The_Combiner: request not recognized");
         }
      }

   public static int[] offset(int ow, int oh, int iw, int ih, int where) {
      return new int[]{ (new int[]{0, (ow-iw)/2, ow-iw})[where%3], (new int[]{0, (oh-ih)/2, oh-ih})[where/3] };
      }

   public static ImagePlus combine(ImagePlus[][] imps) {
      return combine(imps,0); 
      }
   public static ImagePlus combine(ImagePlus[][] imps, int where) {
      ImagePlus[] himps = new ImagePlus[imps.length];
      for(int r=0; r<imps.length; r++)
         himps[r] = combineHorizontally(imps[r],where);
      return combineVertically(himps,where); //null; // TBC
      }


   public static ImagePlus combineHorizontally(Roi roi, int bx, int by, String... titles) {
      ImagePlus[] imps = new ImagePlus[titles.length];
      for(int i=0; i<titles.length; i++) {
         imps[i] = WindowManager.getImage(titles[i]).duplicate();
         IJ.run(imps[i], "RGB Color", "");
         Dimension size = new Dimension(imps[i].getWidth()+2*bx, imps[i].getHeight()+2*by);
         if (roi != null) {
            imps[i].setRoi(roi);
            IJ.run(imps[i], "Clear Outside", "");
            imps[i] = imps[i].crop();
            Rectangle r = roi.getBounds();
            r.grow(2*bx,2*by);
            size = r.getSize();
            }
         IJ.run(imps[i], "Canvas Size...", "width="+size.width+" height="+size.height+" position=Center zero");
         }
      return combineHorizontally(imps);
      }
   public static ImagePlus combineHorizontally(String... titles) {
      ImagePlus[] imps = new ImagePlus[titles.length];
      for(int i=0; i<titles.length; i++) {
         imps[i] = WindowManager.getImage(titles[i]).duplicate();
         IJ.run(imps[i], "RGB Color", "");
         }
      return combineHorizontally(imps);
      }
   public static ImagePlus combineHorizontally(ImagePlus... imps) {
      return combineHorizontally(imps,0);
      }
   public static ImagePlus combineHorizontally(ImagePlus[] imps, int where) {
      if (imps.length < 1) return null;

      int oh=0, ow=0; 
      for(int i=0; i<imps.length; i++) {
         if (imps[i] == null) continue;
         oh = Math.max(oh,imps[i].getHeight());
         ow += imps[i].getWidth();
         }
      
      ImageStack mis = new ImageStack(ow,oh);
      for(int iz=0; iz<imps[0].getImageStackSize(); iz++) {
         ImageProcessor mip = imps[0].getProcessor().createProcessor(ow,oh);
         mip.setColor(Toolbar.getBackgroundColor());
         mip.fill();
         for(int i=0,ws=0; i<imps.length; i++) {
            if (imps[i] == null) continue;
            int[] x0y0 = offset(imps[i].getWidth(),oh,imps[i].getWidth(),imps[i].getHeight(),where);
            mip.insert(imps[i].getStack().getProcessor(iz+1),ws,x0y0[1]);
            ws+=imps[i].getWidth();
            }
         mis.addSlice(mip);
         }
      ImagePlus mimp = new ImagePlus(null, mis);
      //mimp.copyAttributes(imps[0]);
      mimp.setDisplayRange(imps[0].getDisplayRangeMin(),imps[0].getDisplayRangeMax());
      return mimp;
      }
   public static ImagePlus combineVertically(ImagePlus... imps) {
      return combineVertically(imps,0);
      }
   public static ImagePlus combineVertically(ImagePlus[] imps, int where) {
      if (imps.length < 1) return null;
      int oh=0, ow=0; 
      for(int i=0; i<imps.length; i++) 
         if (imps[i] != null) {
            ow = Math.max(ow,imps[i].getWidth());
            oh += imps[i].getHeight();
            }
      
      ImageStack mis = new ImageStack(ow,oh);
      for(int iz=0; iz<imps[0].getImageStackSize(); iz++) {
         ImageProcessor mip = imps[0].getProcessor().createProcessor(ow,oh);
         mip.setColor(Toolbar.getBackgroundColor());
         mip.fill();
         for(int i=0,hs=0; i<imps.length; hs+=imps[i].getHeight(),i++) {
            int[] x0y0 = offset(ow,imps[i].getHeight(),imps[i].getWidth(),imps[i].getHeight(),where);
            mip.insert(imps[i].getStack().getProcessor(iz+1),x0y0[0],hs);
            }
         mis.addSlice(mip);
         }
      ImagePlus mimp = new ImagePlus(null, mis);
      //mimp.copyAttributes(imps[0]);
      mimp.setDisplayRange(imps[0].getDisplayRangeMin(),imps[0].getDisplayRangeMax());
      return mimp;
      }

   public static ImagePlus montage(ImagePlus[][] imps) {
      return montage(imps,0);
      }
   public static ImagePlus montage(ImagePlus[][] imps, int where) {
      int h=0, w=0; 
      for(int ih=0; ih<imps.length; ih++) 
         for(int iw=0; iw<imps[ih].length; iw++) 
            if (imps[ih][iw] != null) {
               h = Math.max(h,imps[ih][iw].getHeight());
               w = Math.max(w,imps[ih][iw].getWidth());
               }
      return montage(imps,w,h,where);
      }
   public static ImagePlus montage(ImagePlus[][] imps, int w, int h, int where) { 
      ImagePlus[] impsv = new ImagePlus[imps.length];
      for(int i=0; i<imps.length; i++)
         impsv[i] = montageHorizontally(imps[i], w, h, where);
      return montageVertically(impsv, imps[0].length*w, h, where);
      }

   public static ImagePlus montageHorizontally(ImagePlus... imps) {
      if (imps.length < 1) return null;

      int oh=0, ow=0; 
      for(int i=0; i<imps.length; i++) 
         if (imps[i] != null) {
            ow = Math.max(ow,imps[i].getWidth());
            oh = Math.max(oh,imps[i].getHeight());
            }

      return montageHorizontally(imps, ow, oh, 0);
      }
   public static ImagePlus montageHorizontally(ImagePlus[] imps, int width, int height, int where) {
      if (imps.length < 1) return null;

      int ss = 0;
      for(ImagePlus imp : imps)
         if (imp != null) 
            ss = Math.max(ss,imp.getImageStackSize());

      ImageStack mis = new ImageStack(imps.length*width,height);
      for(int iz=0; iz<ss; iz++) {
         ImageProcessor mip = imps[0].getProcessor().createProcessor(width,imps.length*height);
         mip.setColor(Toolbar.getBackgroundColor());
         mip.fill();
         for(int i=0; i<imps.length; i++) 
            if (imps[i] != null) {
               int[] x0y0 = offset(width,height,imps[i].getWidth(),imps[i].getHeight(),where);
               mip.insert(imps[i].getStack().getProcessor(iz+1),i*width+x0y0[0],x0y0[1]);
               }
         mis.addSlice(mip);
         }
      ImagePlus mimp = new ImagePlus("Montage", mis);
      //mimp.copyAttributes(imps[0]);
      //mimp.setDisplayRange(imps[0].getDisplayRangeMin(),imps[0].getDisplayRangeMax());
      return mimp;
      }
   public static ImagePlus montageVertically(ImagePlus... imps) {
      if (imps.length < 1) return null;
      return montageVertically(imps, imps[0].getWidth(), imps[0].getHeight(), 0);
      }
   public static ImagePlus montageVertically(ImagePlus[] imps, int width, int height, int where) {
      if (imps.length < 1) return null;

      int ss = 0;
      for(ImagePlus imp : imps)
         if (imp != null) 
            ss = Math.max(ss,imp.getImageStackSize());

      ImageStack mis = new ImageStack(width,imps.length*height);
      for(int iz=0; iz<ss; iz++) {
         ImageProcessor mip = imps[0].getProcessor().createProcessor(width,imps.length*height);
         mip.setColor(Toolbar.getBackgroundColor());
         mip.fill();
         for(int i=0; i<imps.length; i++) 
            if (imps[i] != null) {
               int[] x0y0 = offset(width,height,imps[i].getWidth(),imps[i].getHeight(),where);
               //mip.insert(imps[i].getStack().getProcessor(iz+1),x0y0[0],i*height+x0y0[1]);
               ImageProcessor ip = imps[i].getStack().getProcessor(iz+1);
               ip.setMinAndMax(imps[0].getDisplayRangeMin(),imps[0].getDisplayRangeMax());
               mip.insert(ip,x0y0[0],i*height+x0y0[1]);
               }
         mis.addSlice(mip);
         }
      ImagePlus mimp = new ImagePlus(null, mis);
      //mimp.copyAttributes(imps[0]);
      mimp.setDisplayRange(imps[0].getDisplayRangeMin(),imps[0].getDisplayRangeMax());
      return mimp;
      }

   public static ImagePlus montage(ImagePlus imp, String var) {
      TreeMap<Double,ImagePlus> map = F_Project.byLabel(imp, var);
      Vector<ImagePlus> vimp = new Vector<ImagePlus>();
      for(Map.Entry<Double,ImagePlus> e : map.entrySet()) {
         vimp.add(montageHorizontally(new ImagePlus[]{e.getValue()}));
         }
      return montageVertically(vimp.toArray(new ImagePlus[0]));
      }
   public static ImagePlus montageWL(ImagePlus imp, String var) {
      TreeMap<Double,ImagePlus> map = F_Project.byLabel(imp, var);
      Vector<ImagePlus> vimp = new Vector<ImagePlus>();
      for(Map.Entry<Double,ImagePlus> e : map.entrySet()) {
         vimp.add(montageHorizontally(e.getValue()));
         }
      return montageVertically(vimp.toArray(new ImagePlus[0]));
      }


   public static ImagePlus montageHorizontally(ImagePlus imp) {
      int[][] ind = new int[1][imp.getNFrames()];
      for(int i=0; i<imp.getNFrames(); i++)
         ind[0][i] = i+1;
      return montage(imp,ind);
      }
   public static ImagePlus montageVertically(ImagePlus imp) {
      int[][] ind = new int[imp.getNFrames()][1];
      for(int i=0; i<imp.getNFrames(); i++)
         ind[i][0] = i+1;
      return montage(imp,ind);
      }
   public static ImagePlus montage(ImagePlus imp, int[][] ind) {
      int w = imp.getWidth();
      int h = imp.getHeight();
      ImageStack mis = new ImageStack(w*ind[0].length,h*ind.length);
      for(int iz=0; iz<imp.getNSlices(); iz++) {
         ImageProcessor mip = imp.getProcessor().createProcessor(w*ind[0].length,h*ind.length).convertToRGB();
         for(int iw=0; iw<ind[0].length; iw++)
            for(int ih=0; ih<ind.length; ih++)
               if (ind[ih][iw] > 0) {
                  ImageProcessor ip = imp.getStack().getProcessor(imp.getStackIndex(1,iz+1,ind[ih][iw])).convertToFloat();
                  ip.resetMinAndMax();
                  mip.insert(ip.convertToRGB(),iw*w,ih*h);
                  //mip.insert(imp.getStack().getProcessor(imp.getStackIndex(1,iz+1,ind[ih][iw])),iw*w,ih*h);
                  }
         mip.resetMinAndMax();
         mis.addSlice(mip);
         }

      ImagePlus mimp = new ImagePlus(null, mis);
//      mimp.copyAttributes(imp);
      mimp.setDisplayRange(imp.getDisplayRangeMin(),imp.getDisplayRangeMax());
      return mimp;
      }
}
